//! The third challenge, in which the contestant must answer a series of math questions within a time limit.
//!
//! This challenge requires the contestant to establish a TCP connection to the server and answer a
//! series of math questions that are randomly generated by the server. The contestant must answer
//! [NEEDED_CORRECT] questions correctly within a time limit of [MILLIS_PER_QUESTION] milliseconds
//! per question. If the contestant fails to answer a question correctly or runs out of time, their
//! counter of correct answers will reset.
//!
//! The math questions are generated randomly and have the form "What is x \[op\] y?", where \[op\] is one
//! of the four basic arithmetic operations (+, -, *, /) and x and y are random integers between
//! 0 and 2^16-1.
//!
//! The contestant must send their answers as bytes (not text) over the TCP connection.
//!
//! It is expected that no human can calculate the questions fast enough, so scripting a math
//! solver is needed.
//!
//! See the [sample solution for c3](https://git.cscherr.de/PlexSheep/wooly-vault/src/branch/master/solutions/3.py).

use std::fmt::Display;
use std::net::SocketAddr;

use anyhow::Result;
use async_trait::async_trait;
use libpt::log::{debug, info, warn};
use rand::distributions::{Distribution, Standard};
use rand::{random, Rng};
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::net::{TcpListener, TcpStream};
use tokio::time::Instant;

use super::Challenge;
use crate::config::Config;
use crate::has_won;
use crate::vault::VaultRef;

/// Amount of questions required in the time limit to win.
pub const NEEDED_CORRECT: usize = 16;
/// Amount of milliseconds allotted for each question, used to calculate the time limit.
pub const MILLIS_PER_QUESTION: u128 = 200;

/// Signifies the operations that the randomiser can come up with.
#[derive(Copy, Clone)]
enum Operation {
    Add,
    Sub,
    Div,
    Mul,
}

impl Operation {
    /// Calculate the result of two [i32] with this [Operation].
    ///
    /// This will not error from overflows.
    pub fn calc(self, a: i32, b: i32) -> i32 {
        match self {
            Self::Add => a.overflowing_add(b).0,
            Self::Sub => a.overflowing_sub(b).0,
            Self::Div => a.overflowing_div(b).0,
            Self::Mul => a.overflowing_mul(b).0,
        }
    }
}

impl Display for Operation {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Self::Add => "+",
                Self::Sub => "-",
                Self::Mul => "*",
                Self::Div => "/",
            }
        )
    }
}

impl Distribution<Operation> for Standard {
    fn sample<R: rand::Rng + ?Sized>(&self, rng: &mut R) -> Operation {
        match rng.gen_range(0..3) {
            0 => Operation::Add,
            1 => Operation::Sub,
            2 => Operation::Mul,
            3 => Operation::Div,
            _ => unreachable!(),
        }
    }
}

/// This struct holds the configuration and vault for the challenge.
pub struct C3 {
    config: Config,
    vault: VaultRef,
}

impl C3 {
    /// Create a new randomized math question.
    ///
    /// Questions will always have the same structure:
    ///
    /// * "What is x + y?" (where the operation can be any of [Operation].)
    ///
    /// `x` and `y` will always be a random integer between 0 and 2^16-1.
    ///
    /// # Returns
    ///
    /// This function returns both the question and the solution to that question as a tuple. The
    /// first is the question and the second is the solution.
    pub fn mk_question() -> (String, String) {
        let mut rng = rand::thread_rng();
        let a: i32 = rng.gen_range(0i32..i16::MAX as i32);
        let b: i32 = rng.gen_range(0i32..i16::MAX as i32);
        let op: Operation = random();
        let res = op.calc(a, b);

        let question = format!("What is {a} {op} {b}?");
        let solution = format!("{}", res);
        (question, solution)
    }

    /// Handle a singular contestant
    ///
    /// Messages sent by Wooly Vault [C3] will always be terminated with a newline. [C3] will
    /// [randomly generate questions](Self::mk_question) that the contestant needs to solve in
    /// a given time limit. If the contestant fails to correctly answer [NEEDED_CORRECT] questions
    /// in the given time limit, their counter of correct answer will reset.
    ///
    /// The time limit is [the amount of required correct answers](NEEDED_CORRECT) times [the
    /// milliseconds per question](MILLIS_PER_QUESTION).
    ///
    /// # Errors
    ///
    /// This function will return an error if using the [TcpStream] fails.
    pub async fn handler(peer: (TcpStream, SocketAddr), vault: &VaultRef) -> Result<()> {
        let mut stream = peer.0;
        let addr = peer.1;
        let mut correct: usize = 0;

        let mut buf: Vec<u8> = Vec::new();

        let (stream_read, mut stream_write) = stream.split();
        let mut rdbuf = BufReader::new(stream_read);

        let mut last_reset = Instant::now();

        loop {
            if correct >= NEEDED_CORRECT {
                Self::win(vault, &mut stream, &addr).await?;
                break;
            }
            if last_reset.elapsed().as_millis() > NEEDED_CORRECT as u128 * MILLIS_PER_QUESTION {
                correct = 0;
                debug!("{addr} was too slow, resetting");
                stream_write
                    .write_all("You're too slow!\n".to_string().as_bytes())
                    .await?;
                last_reset = Instant::now();
                continue;
            }
            let (question, solution) = Self::mk_question();
            debug!("sending question to {addr}: '{question}'");
            stream_write.write_all((question + "\n").as_bytes()).await?;
            rdbuf.read_until(0x0A, &mut buf).await?; // read until a newline comes
            let answer = String::from_utf8_lossy(&buf).trim().to_string();
            if answer == solution {
                correct += 1;
                debug!("{addr} got one correct");
                stream_write
                    .write_all(format!("correct! {correct} / {NEEDED_CORRECT}\n").as_bytes())
                    .await?;
            } else {
                debug!("{addr} screwed up, {answer} != {solution}");
                stream_write
                    .write_all("wrong!\n".to_string().as_bytes())
                    .await?;
                correct = 0;
                last_reset = Instant::now();
            }
            buf.clear()
        }

        Ok(())
    }

    async fn win(
        vault: &VaultRef,
        stream: &mut TcpStream,
        addr: &std::net::SocketAddr,
    ) -> Result<()> {
        has_won(addr);
        if let Err(e) = stream
            .write_all(("You win: ".to_owned() + vault.secret()).as_bytes())
            .await
        {
            warn!("could not write to peer {addr}: {e}");
            return Err(e.into());
        };
        if let Err(e) = stream.shutdown().await {
            warn!("could end connection to peer {addr}: {e}");
            return Err(e.into());
        };
        Ok(())
    }
}

#[async_trait]
impl Challenge for C3 {
    fn config(&self) -> Config {
        self.config.clone()
    }
    fn vault(&self) -> VaultRef {
        self.vault.clone()
    }
    fn new(config: Config, vault: VaultRef) -> Self {
        info!("Solution: {}", Self::solution());
        Self { config, vault }
    }

    fn hints() -> Vec<String> {
        vec![
            "TCP connect to 1337 and answer the questions".to_string(),
            "The questions keep chaning".to_string(),
            "You should try to solve the questions in an automated way".to_string(),
        ]
    }

    fn solution() -> String {
        String::from("Connect by TCP, send 1337 as bytes (not text).")
    }

    async fn serve(self) -> anyhow::Result<()> {
        info!("serving challenge 3");
        let listener = TcpListener::bind(self.config.addr).await?;

        loop {
            let vault = self.vault.clone();
            let peer = match listener.accept().await {
                Ok(s) => s,
                Err(err) => {
                    warn!("could not accept tcp stream: {err:?}");
                    continue;
                }
            };
            info!("new peer: {}", peer.1);
            tokio::spawn(async move {
                if let Err(e) = Self::handler(peer, &vault).await {
                    warn!("error while handling peer: {e}");
                }
            });
        }
    }
}
